/* BBTrayIcon - a BB4W plugin to add the icon to the system tray.
   Copyright (C) 2005 kana <nicht AT s8 DOT xrea DOT con>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include <assert.h>
#include <stdarg.h>
#include <stdlib.h>

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <shellapi.h>

#include <BBApi-C.h>
#include <BBCmd.h>


#ifdef __GNUC__
#define PRINTF_LIKE(string_index, first_to_check) \
        __attribute__((format(printf, string_index, first_to_check)))
#else
#define PRINTF_LIKE(string_index, first_to_check)
#endif

#define NUMBER_OF(array) (sizeof(array) / sizeof((array)[0]))
#define MIN(a,b) ((a) <= (b) ? (a) : (b))
#define IS_NULL_STRING(s) ((s)[0] == '\0')




/* Constants */

static const char BBTI_NAME[] = "BBTrayIcon";
static const char BBTI_VERSION[] = "0.0.1";
static const char BBTI_AUTHOR[] = "kana";
static const char BBTI_RELEASE[] = "2005-01-21/25";
static const char BBTI_LINK[] = "http://nicht.s8.xrea.com/2005/01/BBTrayIcon";
static const char BBTI_EMAIL[] = "nicht AT s8 DOT xrea DOT con";

static const char BBTI_CLASS_NAME[] = "BBTrayIcon";
static const char BBTI_RC_NAME[] = "BBTrayIcon.rc";
static const UINT BBTI_MESSAGES[] = {BB_BROADCAST, BB_RECONFIGURE, 0};

#define SNI_ID 0
#define SNI_CALLBACK_MESSAGE (WM_APP + 0)

enum {
	FIRST_BUTTON,
	LEFT_BUTTON = FIRST_BUTTON,
	RIGHT_BUTTON,
	MIDDLE_BUTTON,
	LAST_BUTTON,
	NUM_OF_BUTTONS = LAST_BUTTON
};

enum {
	FIRST_ACTION,
	ON_DOWN = FIRST_ACTION,
	ON_UP,
	ON_DOUBLECLICK,
	LAST_ACTION,
	NUM_OF_ACTIONS = LAST_ACTION
};




/* Global variables */

static HWND G_PluginWnd = NULL;
static HMODULE G_PluginModule = NULL;
static HICON G_DefaultIcon;


typedef struct _TrayIcon TrayIcon;
struct _TrayIcon {
	char* name;
	NOTIFYICONDATA nid;
	BBCmdCommand* script[NUM_OF_BUTTONS][NUM_OF_ACTIONS];
	BOOL visible;
	TrayIcon* next;
};

typedef TrayIcon TrayIconList;

static TrayIconList* G_TrayIconList;








/* Misc. functions */

static void
VSNPRINTF(char* buf, int size, const char* format, va_list va)
{
	if (buf == NULL || size <= 0)
		return;

	vsnprintf(buf, size, format, va);

	buf[size - 1] = '\0';
}


static char*
STRNCPY(char* dest, const char* src, size_t n)
{
	strncpy(dest, src, n);
	dest[n - 1] = '\0';

	return dest;
}


static char*
strndup(const char* s, int n)
{
	char* p;

	p = malloc(sizeof(char) * (n+1));
	if (p != NULL)
		STRNCPY(p, s, n+1);

	return p;
}


/* BUGS: Doesn't strip backslash-escaped chars in "...". */
typedef struct {
	const char* string;
	int length;
} Token;

static void
tokenize( const char* string,
          Token token[],
          int* token_num,
          int max,
          const char** rest )
{
	const char* WS = " \t";  /* WhiteSpace */
	const char* head;
	const char* tail;
	int n;

	head = string + strspn(string, WS);
	n = 0;
	while ((*head != '\0') && (n < max)) {
		head += strspn(head, WS);

		if (*head == '\'') {
			head++;
			tail = head + strcspn(head, "'");
		} else if (*head == '"') {
			head++;
			tail = head;
			while ((*tail != '\0') && (*tail != '"')) {
				if ((tail[0] == '\\') && (tail[1] != '\0'))
					tail++;
				tail++;
			}
		} else {
			tail = head + strcspn(head, WS);
		}

		token[n].string = head;
		token[n].length = (tail - head) / sizeof(char);

		head = tail + ((*tail == '\'') || (*tail == '"') ? 1 : 0);
		n++;
	}

	*token_num = n;
	*rest = head + strspn(head, WS);
}




PRINTF_LIKE(1, 2)
static void
notice(const char* format, ...)
{
	char buf[80 * 25];
	va_list va;

	va_start(va, format);
	VSNPRINTF(buf, NUMBER_OF(buf), format, va);
	va_end(va);

	MessageBox( G_PluginWnd, buf, BBTI_NAME,
	            MB_OK | MB_SETFOREGROUND | MB_TOPMOST);
}




static BOOL
is_2k_or_later(void)
{
	OSVERSIONINFO os;

	os.dwOSVersionInfoSize = sizeof(os);
	GetVersionEx(&os);

	return 5 <= os.dwMajorVersion;
}




static BOOL
get_module_dir(char* buf, int size)
{
	char* bs;

	if (!GetModuleFileName(G_PluginModule, buf, size))
		return FALSE;
	buf[size - 1] = '\0';

	bs = strrchr(buf, '\\');
	if (bs != NULL)
		bs[1] = '\0';

	return TRUE;
}


static BOOL
get_rc_path(char* buf, int size, const char* rc_name)
{
	char dir[MAX_PATH];

	if (!get_module_dir(dir, NUMBER_OF(dir)))
		return FALSE;

	if (size < strlen(dir) + strlen(rc_name) + 1)
		return FALSE;

	strcpy(buf, dir);
	strcat(buf, rc_name);

	return TRUE;
}




static int
convert_button(Token token)
{
	if (!strnicmp("left", token.string, token.length))
		return LEFT_BUTTON;
	else if (!strnicmp("right", token.string, token.length))
		return RIGHT_BUTTON;
	else  /* middle */
		return MIDDLE_BUTTON;
}


static int
convert_action(Token token)
{
	if (!strnicmp("down", token.string, token.length))
		return ON_DOWN;
	else if (!strnicmp("up", token.string, token.length))
		return ON_UP;
	else  /* double-click */
		return ON_DOUBLECLICK;
}




static HICON
load_icon(Token path)
{
	HICON hicon;
	int index;
	char* p;
	char tmp[MAX_PATH];

	STRNCPY(tmp, path.string, MIN(NUMBER_OF(tmp), path.length+1));

	p = strrchr(tmp, ':');
	if ((p != NULL) && (strlen(p+1) == strspn(p+1, "-0123456789"))) {
		index = atoi(p+1);
		if (index < 0)
			index = -index;
		*p = '\0';
	} else {
		index = 0;
	}

	hicon = ExtractIcon(G_PluginModule, tmp, index);

	return (hicon != NULL) ? hicon : G_DefaultIcon;
}




static TrayIcon*
TrayIcon_New(Token name, Token icon_path, Token tooltip)
{
	TrayIcon* self;

	self = malloc(sizeof(*self));
	if (self == NULL)
		return NULL;

	self->name = strndup(name.string, name.length);
	if (self->name == NULL) {
		free(self);
		return NULL;
	}

	self->visible = TRUE;
	self->next = NULL;
	ZeroMemory(self->script, sizeof(self->script));

	self->nid.cbSize = sizeof(self->nid);
	self->nid.hWnd = G_PluginWnd;
	self->nid.uID = (UINT)self;
	self->nid.uFlags = NIF_ICON | NIF_MESSAGE | NIF_TIP;
	self->nid.uCallbackMessage = SNI_CALLBACK_MESSAGE;
	self->nid.hIcon = load_icon(icon_path);
	STRNCPY( self->nid.szTip, tooltip.string,
	         MIN(NUMBER_OF(self->nid.szTip), tooltip.length+1) );

	Shell_NotifyIcon(NIM_ADD, &(self->nid));
	return self;
}


static void
TrayIcon_Del(TrayIcon* self)
{
	int button;
	int action;

	if (self->visible)
		Shell_NotifyIcon(NIM_DELETE, &(self->nid));

	for (button = FIRST_BUTTON; button < LAST_BUTTON; button++) {
		for (action = FIRST_ACTION; action < LAST_ACTION; action++) {
			BBCmdCommand_Del(self->script[button][action]);
		}
	}
	if (self->nid.hIcon != G_DefaultIcon)
		DestroyIcon(self->nid.hIcon);
	free(self->name);
	free(self);
}


static int
TrayIcon_SetIcon(TrayIcon* target, void* param)
{
	HICON old_icon;

	old_icon = target->nid.hIcon;

	target->nid.hIcon = load_icon(*(Token*)param);
	if (target->visible)
		Shell_NotifyIcon(NIM_MODIFY, &(target->nid));

	if ((old_icon != NULL) && (old_icon != G_DefaultIcon))
		DestroyIcon(old_icon);
	return NO_ERROR;
}

static int
TrayIcon_SetTooltip(TrayIcon* target, void* param)
{
	STRNCPY(
	  target->nid.szTip, ((Token*)param)->string,
	  MIN(NUMBER_OF(target->nid.szTip), ((Token*)param)->length+1)
	);
	if (target->visible)
		Shell_NotifyIcon(NIM_MODIFY, &(target->nid));
	return NO_ERROR;
}

static int
TrayIcon_SetScript(TrayIcon* target, void* param)
{
	BBCmdCompiler* compiler;
	BBCmdCommand* cmd;
	int line;
	int column;
	const char* message;
	Token* token;
	int button;
	int action;
	const char* script;

	/* 0:button, 1:action, 2:name, 3/rest:script */
	token = param;
	script = token[3].string;
	button = convert_button(token[0]);
	action = convert_action(token[1]);

	if (IS_NULL_STRING(script)) {
		BBCmdCommand_Del(target->script[button][action]);
		target->script[button][action] = NULL;
		return NO_ERROR;
	}

	compiler = BBCmdCompiler_New();
	if (compiler == NULL)
		return ERROR_OUTOFMEMORY;

	cmd = BBCmdCompiler_CompileFromString( compiler, script,
	                                       &line, &column, &message );
	if (cmd == NULL) {
		notice( "Compile error\n"
		        "\n"
		        "%s\n"
		        "Line %d, column %d: %s",
		        token[0].string,
		        line, column, message );
		goto E_COMPILE;
	}

	BBCmdCommand_Del(target->script[button][action]);
	target->script[button][action] = cmd;

E_COMPILE:
	BBCmdCompiler_Del(compiler);
	return NO_ERROR;
}

static int
TrayIcon_Show(TrayIcon* target, void* param)
{
	if (!(target->visible)) {
		Shell_NotifyIcon(NIM_ADD, &(target->nid));
		target->visible = TRUE;
	}
	return NO_ERROR;
}

static int
TrayIcon_Hide(TrayIcon* target, void* param)
{
	if (target->visible) {
		Shell_NotifyIcon(NIM_DELETE, &(target->nid));
		target->visible = FALSE;
	}
	return NO_ERROR;
}

static int
TrayIcon_Toggle(TrayIcon* target, void* param)
{
	target->visible = !(target->visible);
	Shell_NotifyIcon( (target->visible ? NIM_ADD : NIM_DELETE),
	                  &(target->nid) );
	return NO_ERROR;
}

static int
TrayIcon_Info(TrayIcon* target, void* unused)
{
	notice( "Name:    %s\n"
	        "Visible: %s\n"
	        "Script:  on %c%c%c down, %c%c%c up, %c%c%c double-click\n"
	        "Icon:    %s\n",
	        target->name,
	        (target->visible ? "yes" : "no"),
	        (target->script[LEFT_BUTTON][ON_DOWN] ? 'L' : '_'),
	        (target->script[RIGHT_BUTTON][ON_DOWN] ? 'R' : '_'),
	        (target->script[MIDDLE_BUTTON][ON_DOWN] ? 'M' : '_'),
	        (target->script[LEFT_BUTTON][ON_UP] ? 'L' : '_'),
	        (target->script[RIGHT_BUTTON][ON_UP] ? 'R' : '_'),
	        (target->script[MIDDLE_BUTTON][ON_UP] ? 'M' : '_'),
	        (target->script[LEFT_BUTTON][ON_DOUBLECLICK] ? 'L' : '_'),
	        (target->script[RIGHT_BUTTON][ON_DOUBLECLICK] ? 'R' : '_'),
	        (target->script[MIDDLE_BUTTON][ON_DOUBLECLICK] ? 'M' : '_'),
	        (target->nid.hIcon==G_DefaultIcon ? "Default" : "Original") );
	return NO_ERROR;
}




static TrayIconList*
TrayIconList_New(void)
{
	return NULL;
}


static void
TrayIconList_Del(TrayIconList* self)
{
	TrayIcon* target;
	TrayIcon* next;

	target = self;
	while (target != NULL) {
		next = target->next;
		TrayIcon_Del(target);
		target = next;
	}
}


static int
TrayIconList_Add( TrayIconList** self,
                  Token name, Token icon_path, Token tooltip )
{
	TrayIcon* trayicon;

	trayicon = TrayIcon_New(name, icon_path, tooltip);
	if (trayicon == NULL)
		return ERROR_OUTOFMEMORY;

	if (*self == NULL)
		*self = trayicon;
	else {
		TrayIcon* t;

		t = *self;
		while (TRUE) {
			if (!strncmp(t->name, name.string, name.length)) {
				TrayIcon_Del(trayicon);
				return ERROR_ALREADY_EXISTS;
			}

			if (t->next == NULL)
				break;
			t = t->next;
		}

		t->next = trayicon;
	}
	return NO_ERROR;
}


static BOOL
TrayIconList_DelByName(TrayIconList** self, Token name)
{
	TrayIcon* target;
	TrayIcon* prev;

	prev = NULL;
	target = *self;
	while (target != NULL) {
		if (!strncmp(target->name, name.string, name.length)) {
			if (prev == NULL)
				*self = target->next;
			else
				prev->next = target->next;
			TrayIcon_Del(target);
			return TRUE;
		}

		prev = target;
		target = target->next;
	}

	return FALSE;
}



typedef int (*TrayIcon_Process_f)(TrayIcon*, void*);

static int
TrayIconList_SearchAndProcess( TrayIconList* self,
                               Token name,
                               TrayIcon_Process_f process,
                               void* process_param )
{
	TrayIcon* target;

	target = self;
	while (target != NULL) {
		if (!strncmp(target->name, name.string, name.length))
			return (*process)(target, process_param);

		target = target->next;
	}

	return ERROR_NOT_FOUND;
}








/* Main functions */

static BOOL
read_settings(void)
{
	char rc[MAX_PATH];
	FILE* file;
	char line[256];
	char* nl;

	if (!get_rc_path(rc, NUMBER_OF(rc), BBTI_RC_NAME)) {
		notice("Failed to load %s (too long path)", BBTI_RC_NAME);
		return FALSE;
	}

	file = fopen(rc, "rt");
	if (file == NULL) 
		return TRUE;

	while (fgets(line, NUMBER_OF(line), file) != NULL) {
		if (line[0] != '@')
			continue;

		nl = strchr(line, '\n');
		if (nl != NULL)
			*nl = '\0';

		SendMessage(GetBBWnd(), BB_BROADCAST, 0, (LPARAM)line);
	}

	fclose(file);
	return TRUE;
}




static BOOL
tokenize_error_check( int min, int max, BOOL allow_rest,
                      int token_num, const char* rest,
                      const char* broam )
{
	if ((min == max) && (token_num != min)) {
		notice( "%d arg%s required for %s",
		        min, (min == 1 ? "" : "s"), broam );
		return FALSE;
	}

	if ((token_num < min) || (max < token_num)) {
		notice( "%d or %d arg%s required for %s",
		        min, max, (1 < max ? "s" : ""), broam );
		return FALSE;
	}

	if ((!allow_rest) && (!IS_NULL_STRING(rest))) {
		notice("Trailing characters: %s", broam);
		return FALSE;
	}

	return TRUE;
}


void
interpret_broam(const char* broam)
{
	Token token[8];
	int token_num;
	const char* rest;
	char tmp[80];

	#define IS(s) (!strncmp((s), token[1].string, token[1].length))

	/* 0:@BBTrayIcon, 1:CMD, rest:ARGS */
	tokenize(broam, token, &token_num, 2, &rest);
	if (token_num != 2)
		return;
	if (strncmp("@BBTrayIcon", token[0].string, token[0].length))
		return;

	if (IS("reload")) {
		PostMessage(G_PluginWnd, BB_RECONFIGURE, 0, 0);
	} else if (IS("clear")) {
		TrayIconList_Del(G_TrayIconList);
		G_TrayIconList = TrayIconList_New();
	} else if (IS("add")) {
		int result;

		/* 0:name, 1:icon, (2:tooltip) */
		tokenize(rest, token, &token_num, 3, &rest);
		if (!tokenize_error_check(2, 3, FALSE, token_num, rest, broam))
			return;
		if (token_num == 2)
			token[2].length = 0;

		result = TrayIconList_Add( &G_TrayIconList,
		                           token[0], token[1], token[2] );
		if (result == NO_ERROR)
			return;
		STRNCPY( tmp, token[0].string,
		         MIN(NUMBER_OF(tmp), token[0].length+1) );
		if (result == ERROR_OUTOFMEMORY) {
			notice("Out of memory to add icon `%s'", tmp);
		} else if (result == ERROR_ALREADY_EXISTS) {
			notice("Icon `%s' already exists", tmp);
		} else {
			notice("Unknown error %d at adding `%s'", result, tmp);
		}
	} else if (IS("del")) {
		/* 0:name */
		tokenize(rest, token, &token_num, 1, &rest);
		if (!tokenize_error_check(1, 1, FALSE, token_num, rest, broam))
			return;

		if (!TrayIconList_DelByName(&G_TrayIconList, token[0])) {
			STRNCPY( tmp, token[0].string,
				 MIN(NUMBER_OF(tmp), token[0].length+1) );
			notice("No icon named `%s'", tmp);
		}
	} else if ( IS("show") || IS("hide") || IS("toggle") || IS("info")
	            || IS("icon") || IS("tooltip") || IS("on") )
	{
		TrayIcon_Process_f process;
		int result;
		int min = 1;
		int max = 1;
		BOOL allow_rest = FALSE;
		void* param = NULL;
		Token* name = &(token[0]);

		switch (token[1].string[0]) {
		default: assert(0); break;
		case 's': process = TrayIcon_Show; break;
		case 'h': process = TrayIcon_Hide; break;
		case 't':
			if (token[1].string[2] == 'g')
				process = TrayIcon_Toggle;
			else {
				process = TrayIcon_SetTooltip;
				min = max = 2;
				param = &(token[1]);
			}
			break;
		case 'i':
			if (token[1].string[1] == 'n')
				process = TrayIcon_Info;
			else {
				process = TrayIcon_SetIcon;
				min = max = 2;
				param = &(token[1]);
			}
			break;
		case 'o':
			process = TrayIcon_SetScript;
			min = max = 3;
			allow_rest = TRUE;
			param = token;
			name = &(token[2]);
			break;
		}

		/* 0:... */
		tokenize(rest, token, &token_num, max, &rest);
		if (!tokenize_error_check( min, max, allow_rest,
		                           token_num, rest, broam ))
		{
			return;
		}

		if (process == TrayIcon_SetScript)
			token[3].string = rest;

		result = TrayIconList_SearchAndProcess
		           (G_TrayIconList, *name, process, param);
		if (result == NO_ERROR)
			return;
		STRNCPY( tmp, name->string,
		         MIN(NUMBER_OF(tmp), name->length+1) );
		if (result == ERROR_NOT_FOUND)
			notice("No icon named `%s'", tmp);
		else if (result == ERROR_OUTOFMEMORY)
			notice("Out of memory: %s", broam);
		else
			notice("Unknown error %d: %s", result, broam);
	} else {
		notice("Unknown bro@m: `%s'", broam);
	}

	#undef IS
}




static LRESULT CALLBACK
BBTI_WndProc(HWND hwnd, UINT msg, WPARAM wp, LPARAM lp)
{
	switch (msg) {
	default:
		break;


	case BB_BROADCAST:
		interpret_broam((const char*)lp);
		break;

	case BB_RECONFIGURE:
		read_settings();
		break;


	case SNI_CALLBACK_MESSAGE: {
		TrayIcon* trayicon = (TrayIcon*)wp;
		BBCmdCommand* script;

		if (lp == WM_LBUTTONDOWN)
		  script = trayicon->script[LEFT_BUTTON][ON_DOWN];
		else if (lp == WM_LBUTTONUP)
		  script = trayicon->script[LEFT_BUTTON][ON_UP];
		else if (lp == WM_LBUTTONDBLCLK)
		  script = trayicon->script[LEFT_BUTTON][ON_DOUBLECLICK];
		else if (lp == WM_RBUTTONDOWN)
		  script = trayicon->script[RIGHT_BUTTON][ON_DOWN];
		else if (lp == WM_RBUTTONUP)
		  script = trayicon->script[RIGHT_BUTTON][ON_UP];
		else if (lp == WM_RBUTTONDBLCLK)
		  script = trayicon->script[RIGHT_BUTTON][ON_DOUBLECLICK];
		else if (lp == WM_MBUTTONDOWN)
		  script = trayicon->script[MIDDLE_BUTTON][ON_DOWN];
		else if (lp == WM_MBUTTONUP)
		  script = trayicon->script[MIDDLE_BUTTON][ON_UP];
		else if (lp == WM_MBUTTONDBLCLK)
		  script = trayicon->script[MIDDLE_BUTTON][ON_DOUBLECLICK];
		else
		  script = NULL;

		if (script != NULL)
			BBCmdCommand_Execute(script);
		} break;


	case WM_CREATE:
		SendMessage( GetBBWnd(), BB_REGISTERMESSAGE,
		             (WPARAM)hwnd, (LPARAM)BBTI_MESSAGES );
		break;

	case WM_DESTROY:
		SendMessage( GetBBWnd(), BB_UNREGISTERMESSAGE,
		             (WPARAM)hwnd, (LPARAM)BBTI_MESSAGES );
		break;
	}

	return DefWindowProc(hwnd, msg, wp, lp);
}








/* Public functions */

DLL_EXPORT int
beginPlugin(HINSTANCE hplugininstance)
{
	WNDCLASS wc;

	if ( BBCmd_IsCompatible(BBCmd_VERSION, BBCmd_ICH_JUST_USE)
	     != BBCmd_IC_COMPATIBLE )
	{
		notice("Imcompatible BBCmd");
		goto E_BBCMD_COMPATIBILITY;
	}

	G_PluginModule = (HMODULE)hplugininstance;
	G_DefaultIcon = LoadIcon(NULL, IDI_ERROR);

	G_TrayIconList = TrayIconList_New();

	ZeroMemory(&wc, sizeof(wc));
	wc.lpfnWndProc = BBTI_WndProc;
	wc.hInstance = hplugininstance;
	wc.lpszClassName = BBTI_CLASS_NAME;
	if (!RegisterClass(&wc)) {
		notice("Failed RegisterClass()");
		goto E_REGISTER_CLASS;
	}

	G_PluginWnd = CreateWindowEx( WS_EX_TOOLWINDOW,
	                              BBTI_CLASS_NAME,
	                              NULL,
	                              0,
	                              0, 0, 0, 0,
	                              (is_2k_or_later() ? HWND_MESSAGE : NULL),
	                              NULL,
	                              hplugininstance,
	                              NULL );
	if (G_PluginWnd == NULL) {
		notice("Failed CreateWindowEx()");
		goto E_CREATE_WINDOW;
	}

	if (!read_settings())
		goto E_READ_SETTINGS;

	return 0;


E_READ_SETTINGS:
	DestroyWindow(G_PluginWnd);
E_CREATE_WINDOW:
	UnregisterClass(BBTI_CLASS_NAME, hplugininstance);
E_REGISTER_CLASS:
E_BBCMD_COMPATIBILITY:
	return !0;
}




DLL_EXPORT void
endPlugin(HINSTANCE hplugininstance)
{
	DestroyWindow(G_PluginWnd);
	UnregisterClass(BBTI_CLASS_NAME, hplugininstance);
	TrayIconList_Del(G_TrayIconList);
}




DLL_EXPORT LPCSTR
pluginInfo(int field)
{
	switch (field) {
	default: /* FALLTHRU */
	case PLUGIN_NAME: return BBTI_NAME;
	case PLUGIN_VERSION: return BBTI_VERSION;
	case PLUGIN_AUTHOR: return BBTI_AUTHOR;
	case PLUGIN_RELEASE: return BBTI_RELEASE;
	case PLUGIN_LINK: return BBTI_LINK;
	case PLUGIN_EMAIL: return BBTI_EMAIL;
	}
}




/* __END__ */
